from simple_media_manager.domain.repository.video import VideoWriteRepository


class DjangoImageWriteRepository(VideoWriteRepository):
    def save(self, file: bytes, name: str):
        pass

    def bulk_save(self, files: list):
        pass

    def delete(self, pk: int):
        pass



from drf_spectacular.utils import extend_schema
from rest_framework import serializers, status
from rest_framework.response import Response
from rest_framework.views import APIView

from simple_media_manager.domain.command.image import BulkUploadImageCommand, DeleteImageCommand, GetImageByIDCommand, \
    RetrieveImagesCommand
from simple_media_manager.infrastructure.common.pagination import LimitOffsetPagination, get_paginated_response
from simple_media_manager.infrastructure.dependency.registry import \
    django_compound_image_repository as DjangoCompoundImageRepository
from simple_media_manager.presentation.apis.serializers.image import BriefImageOutputSerializer, \
    ImageMultipleInputSerializer, \
    ImageOutputSerializer


class UploadImageApi(APIView):

    def __init__(self, **kwargs):
        super(UploadImageApi, self).__init__(**kwargs)
        self.upload_image = BulkUploadImageCommand(repository=DjangoCompoundImageRepository)

    @extend_schema(
        request=ImageMultipleInputSerializer(many=True),
        responses=BriefImageOutputSerializer
    )
    def post(self, request):
        serializer = ImageMultipleInputSerializer(many=True, data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            instance = self.upload_image.handle(files=serializer.validated_data)
            result = BriefImageOutputSerializer(instance, many=True, context={'request': request}).data
            return Response(result)
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class GetImagesApi(APIView):

    def __init__(self, **kwargs):
        super(GetImagesApi, self).__init__(**kwargs)
        self.retrieve_images = RetrieveImagesCommand(repository=DjangoCompoundImageRepository)

    class Pagination(LimitOffsetPagination):
        default_limit = 10

    @extend_schema(
        responses=BriefImageOutputSerializer
    )
    def get(self, request):
        try:
            query = self.retrieve_images.handle()
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)
        return get_paginated_response(
            pagination_class=self.Pagination,
            serializer_class=BriefImageOutputSerializer,
            queryset=query,
            request=request,
            view=self,
        )


class GetImageByIdApi(APIView):

    def __init__(self, **kwargs):
        super(GetImageByIdApi, self).__init__(**kwargs)
        self.get_image_by_id = GetImageByIDCommand(repository=DjangoCompoundImageRepository)

    class IdParameterSerializer(serializers.Serializer):
        id = serializers.IntegerField(required=True, allow_null=False)

    @extend_schema(
        parameters=[IdParameterSerializer],
        responses=ImageOutputSerializer
    )
    def get(self, request):
        """
        image_id -- This is the essential parameter for using this api
        """
        try:
            query = self.get_image_by_id.handle(image_id=request.GET.get('id'))
            result = ImageOutputSerializer(query, context={"request": request}).data
            return Response(result)

        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class DeleteImageByIdApi(APIView):

    def __init__(self, **kwargs):
        super(DeleteImageByIdApi, self).__init__(**kwargs)
        self.delete_image_by_id = DeleteImageCommand(repository=DjangoCompoundImageRepository)

    class IdParameterSerializer(serializers.Serializer):
        id = serializers.IntegerField(required=True, allow_null=False)

    @extend_schema(
        parameters=[IdParameterSerializer]
    )
    def delete(self, request):

        param_serializer = self.IdParameterSerializer(data=request.query_params)
        param_serializer.is_valid(raise_exception=True)
        try:
            self.delete_image_by_id.handle(id=param_serializer.validated_data.get('id'))
            return Response(f"Successfully deleted", status=status.HTTP_200_OK)

        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)



from io import BytesIO

from PIL import Image as PillowImage, ImageOps
from django.core.files import File
from django.core.files.uploadedfile import TemporaryUploadedFile

from simple_media_manager.domain.entities.constants import MINIMUM_HHEIGHT, MINIMUM_HWIDTH, MINIMUM_VHEIGHT, MINIMUM_VWIDTH
from simple_media_manager.domain.entities.enums import ImageSizeOrientation


class ImageProcessingService:

    @classmethod
    def _is_image_resizable(cls, min_width: int, min_height: int, image_width: int, image_height: int) -> bool:
        if image_width > min_width or image_height > min_height:
            return True
        return False

    @classmethod
    def _get_new_size(cls, width: int, height: int, resize_percent: int) -> tuple[int, int]:
        new_size = (
            int(width - width * (resize_percent / 100)),
            int(height - height * (resize_percent / 100))
        )
        return new_size

    @classmethod
    def _get_image_orientation(cls, image: PillowImage):
        if image.width < image.height:
            return ImageSizeOrientation.VERTICAL
        elif image.width > image.height:
            return ImageSizeOrientation.HORIZONTAL
        elif image.width == image.height:
            return ImageSizeOrientation.SQUARE

    @classmethod
    def resize_image(cls, image_file: TemporaryUploadedFile, resize_percent: int) -> File:
        image = PillowImage.open(image_file)
        img_format = image.format
        image = ImageOps.exif_transpose(image)
        should_be_resized = False
        match cls._get_image_orientation(image=image):
            case ImageSizeOrientation.VERTICAL:
                should_be_resized = cls._is_image_resizable(MINIMUM_VWIDTH,
                                                            MINIMUM_VHEIGHT,
                                                            image_width=image.width,
                                                            image_height=image.height)
            case ImageSizeOrientation.HORIZONTAL:
                should_be_resized = cls._is_image_resizable(MINIMUM_HWIDTH,
                                                            MINIMUM_HHEIGHT,
                                                            image_width=image.width,
                                                            image_height=image.height)
            case ImageSizeOrientation.SQUARE:
                should_be_resized = cls._is_image_resizable(MINIMUM_VWIDTH,
                                                            MINIMUM_VWIDTH,
                                                            image_width=image.width,
                                                            image_height=image.height)
            case _:
                raise Exception("Orientation is not recognizable")

        if should_be_resized:
            new_size = cls._get_new_size(image.width, image.height, resize_percent=resize_percent)
            image = image.resize(new_size, resample=4)

        image_io = BytesIO()
        image.save(image_io, format=img_format, quality=60, optimize=True)
        django_image = File(image_io, name=image_file.name)
        return django_image



from rest_framework import status as rest_http_status
from rest_framework.response import Response

from ..data_responses.data_response import Error, Message


class MessageResponse(Response):
    def __init__(self, message: str, status: rest_http_status = rest_http_status.HTTP_200_OK):
        super(MessageResponse, self).__init__(data=Message(message=message).__dict__,
                                              status=status)


class ErrorResponse(Response):
    def __init__(self, exception: Exception, status_code=rest_http_status.HTTP_400_BAD_REQUEST):
        super(ErrorResponse, self).__init__(data=Error(errors=exception.__str__()).__dict__,
                                            status=status_code)



from django.urls import include, path

from simple_media_manager.presentation.apis.v1.image import DeleteImageByIdApi, GetImageByIdApi, GetImagesApi, UploadImageApi

urlpatterns = [

    # Image # TODO must be removed later on
    path("", include('simple_media_manager.presentation.apis.v1.urls', namespace='v1')),
    path("v2/", include('simple_media_manager.presentation.apis.v2.urls', namespace='v2')),

]



from django.db import models
from django.utils import timezone


class File(models.Model):
    name = models.CharField(null=True, max_length=100)
    description = models.CharField(null=True, max_length=255)
    created_at = models.DateTimeField(db_index=True, default=timezone.now)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True



"""
Based on the framework architecture which is CQRS, you are supposed to
import this setting into your main Django settings, so they could be
integrated and migrations of this framework will be recognized
"""

MIGRATION_MODULES = {
    'simple_media_manager': 'simple_media_manager.infrastructure.migrations',
}



from abc import ABC, abstractmethod

from . import AbstractRepository
from ...infrastructure.models.image import Image


class ImageReadRepository(AbstractRepository):

    @abstractmethod
    def get(self, pk: int):
        ...

    @abstractmethod
    def find(self, name: str):
        ...

    @abstractmethod
    def all(self):
        ...


class ImageWriteRepository(AbstractRepository):
    @abstractmethod
    def save(self, file: bytes, name: str):
        ...

    @abstractmethod
    def bulk_save(self, files: list):
        ...

    @abstractmethod
    def delete(self, id: int):
        ...


class ImageRepository(ImageReadRepository, ImageWriteRepository, ABC):
    ...


class CompoundImageRepository(ImageRepository):
    def __init__(self, read_repository: ImageReadRepository, write_repository: ImageWriteRepository):
        self.read_repository = read_repository
        self.write_repository = write_repository

    def get(self, pk: int) -> Image:
        return self.read_repository.get(pk=pk)

    def find(self, name: str):
        return self.read_repository.find(name=name)

    def all(self):
        return self.read_repository.all()

    def save(self, file: bytes, name: str):
        return self.write_repository.save(file=file, name=name)

    def bulk_save(self, files: list):
        return self.write_repository.bulk_save(files=files)

    def delete(self, id: int):
        self.write_repository.delete(id=id)



COMPACT_IMAGE_RESIZE_PERCENT = 70
THUMBNAIL_IMAGE_RESIZE_PERCENT = 50
MINIMUM_IMAGE_SIZE = 1000000  # In Kilobytes
MINIMUM_HWIDTH = 1920
MINIMUM_HHEIGHT = 1080
MINIMUM_VWIDTH = 1080
MINIMUM_VHEIGHT = 1920



# Generated by Django 5.0.6 on 2024-06-26 11:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simple_media_manager', '0003_alter_image_options_rename_image_image_original'),
    ]

    operations = [
        migrations.AlterField(
            model_name='image',
            name='compact',
            field=models.ImageField(null=True, upload_to='media/images/compacts', verbose_name='compact size'),
        ),
        migrations.AlterField(
            model_name='image',
            name='original',
            field=models.ImageField(null=True, upload_to='media/images/originals', verbose_name='original size'),
        ),
        migrations.AlterField(
            model_name='image',
            name='thumbnail',
            field=models.ImageField(null=True, upload_to='media/images/thumbnails', verbose_name='thumbnail size'),
        ),
    ]



from simple_media_manager.domain.command import Command
from simple_media_manager.domain.repository.image import ImageRepository


class BaseImageCommand(Command):
    def __init__(self, repository: ImageRepository):
        self.repository = repository


class GetImageByIDCommand(BaseImageCommand):
    def handle(self, image_id: int):
        return self.repository.get(pk=image_id)


class RetrieveImagesCommand(BaseImageCommand):
    def handle(self):
        return self.repository.all()


class BulkUploadImageCommand(BaseImageCommand):

    def handle(self, files: list[dict]):
        return self.repository.bulk_save(files=files)


class SaveImageCommand(BaseImageCommand):

    def handle(self, file: bytes, name: str):
        return self.repository.save(file=file, name=name)


class DeleteImageCommand(BaseImageCommand):
    def handle(self, id: int):
        self.repository.delete(id=id)



from django.urls import path

from simple_media_manager.presentation.apis.v1.image import DeleteImageByIdApi, GetImageByIdApi, GetImagesApi, UploadImageApi

app_name = 'v1'

urlpatterns = [
    # path('upload-image', UploadImageApi.as_view(), name='upload_image'),
    path('get-images', GetImagesApi.as_view(), name='get_images'),
    path('get-image', GetImageByIdApi.as_view(), name='get_image_by_id'),
    path('delete-image', DeleteImageByIdApi.as_view(), name='delete_image_by_id'),
]



from django.urls import include, path

from simple_media_manager.presentation.apis.v2.image import DeleteImageByIdApi, GetImageByIdApi, SearchImagesApi, \
    UpdateImageApi, \
    UploadImagesApi

app_name = 'v2'
urlpatterns = [
    path('image/', include([
        path('<int:pk>/delete', DeleteImageByIdApi.as_view(), name='delete_image_by_id'),
        path('<int:pk>', GetImageByIdApi.as_view(), name='get_image'),
        path('<int:pk>/update', UpdateImageApi.as_view(), name='upload_image_v2'),
        path('upload', UploadImagesApi.as_view(), name='upload_image_v2'),
        path('search', SearchImagesApi.as_view(), name='search_images'),
    ]))
]



# Generated by Django 4.0.7 on 2024-04-19 10:27

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, null=True)),
                ('description', models.CharField(max_length=255, null=True)),
                ('created_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('image', models.ImageField(null=True, upload_to='media/simple_media_manager/images')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from drf_spectacular.utils import extend_schema
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from simple_media_manager.domain.command.image import BulkUploadImageCommand, DeleteImageCommand, GetImageByIDCommand, \
    RetrieveImagesCommand
from simple_media_manager.infrastructure.common.pagination import LimitOffsetPagination, get_paginated_response
from simple_media_manager.infrastructure.dependency.registry import django_compound_image_repository as DjangoCompoundImageRepository
from simple_media_manager.presentation.apis.v2.serializers.image import BriefImageOutputSerializer, \
    ImageMultipleInputSerializer, \
    ImageOutputSerializer, SearchImageInputParameterSerializer


class UploadImagesApi(APIView):

    def __init__(self, **kwargs):
        super(UploadImagesApi, self).__init__(**kwargs)
        self.upload_image = BulkUploadImageCommand(repository=DjangoCompoundImageRepository)

    @extend_schema(
        request={
            "multipart/form-data": {
                "type": "object",
                "properties": {
                    "images": {
                        "type": "array",
                        "items": {
                            "type": "string",
                            "format": "binary", }}}}},
        responses=BriefImageOutputSerializer
    )
    def post(self, request):
        serializer = ImageMultipleInputSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            instance = self.upload_image.handle(files=serializer.validated_data.get('images'))
            result = BriefImageOutputSerializer(instance, many=True, context={'request': request}).data
            return Response(result)
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class UpdateImageApi(APIView):

    def __init__(self, **kwargs):
        super(UpdateImageApi, self).__init__(**kwargs)
        self.upload_image = BulkUploadImageCommand(repository=DjangoCompoundImageRepository)

    @extend_schema(
        request=ImageMultipleInputSerializer(many=True),
        responses=BriefImageOutputSerializer
    )
    def patch(self, request, pk: int):
        serializer = ImageMultipleInputSerializer(many=True, data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            instance = self.upload_image.handle(files=serializer.validated_data)
            result = BriefImageOutputSerializer(instance, many=True, context={'request': request}).data
            return Response(result)
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class SearchImagesApi(APIView):

    def __init__(self, **kwargs):
        super(SearchImagesApi, self).__init__(**kwargs)
        self.retrieve_images = RetrieveImagesCommand(repository=DjangoCompoundImageRepository)

    class Pagination(LimitOffsetPagination):
        default_limit = 10

    @extend_schema(
        parameters=[SearchImageInputParameterSerializer],
        responses=BriefImageOutputSerializer
    )
    def get(self, request):
        try:
            query = self.retrieve_images.handle()
            return get_paginated_response(
                pagination_class=self.Pagination,
                serializer_class=BriefImageOutputSerializer,
                queryset=query,
                request=request,
                view=self,
            )

        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class GetImageByIdApi(APIView):
    def __init__(self, **kwargs):
        super(GetImageByIdApi, self).__init__(**kwargs)
        self.get_image_by_id = GetImageByIDCommand(repository=DjangoCompoundImageRepository)

    @extend_schema(
        responses=ImageOutputSerializer
    )
    def get(self, request, pk):
        try:
            query = self.get_image_by_id.handle(image_id=pk)
            result = ImageOutputSerializer(query, context={"request": request}).data
            return Response(result)
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)


class DeleteImageByIdApi(APIView):
    def __init__(self, **kwargs):
        super(DeleteImageByIdApi, self).__init__(**kwargs)
        self.delete_image_by_id = DeleteImageCommand(repository=DjangoCompoundImageRepository)

    def delete(self, request, pk):
        try:
            self.delete_image_by_id.handle(id=pk)
            return Response(f"Successfully deleted", status=status.HTTP_200_OK)
        except Exception as e:
            return Response(f"Database Error {e}", status=status.HTTP_400_BAD_REQUEST)



from collections.abc import Iterator

from django.core.files.uploadedfile import TemporaryUploadedFile

from simple_media_manager.domain.entities.constants import COMPACT_IMAGE_RESIZE_PERCENT, THUMBNAIL_IMAGE_RESIZE_PERCENT, \
    MINIMUM_IMAGE_SIZE
from simple_media_manager.domain.repository.image import ImageReadRepository, ImageWriteRepository
from simple_media_manager.infrastructure.models import Image
from simple_media_manager.infrastructure.services.image import ImageProcessingService


class DjangoImageWriteRepository(ImageWriteRepository):

    def save(self,
             file: TemporaryUploadedFile,
             name: str = '',
             image_processing_service: ImageProcessingService = ImageProcessingService()) -> Image:
        return Image.objects.create(name=name,
                                    original=file,
                                    thumbnail=image_processing_service.resize_image(image_file=file, resize_percent=60),
                                    compact=image_processing_service.resize_image(image_file=file, resize_percent=80))

    def create(self,
               file: TemporaryUploadedFile,
               name: str,
               image_processing_service: ImageProcessingService = ImageProcessingService()) -> Image:
        """
        Creates a new Image object, and it isn't saved into database, only held in RAM
        """

        if not file.size < MINIMUM_IMAGE_SIZE:
            instance = Image(name=name,
                             original=file,
                             thumbnail=image_processing_service
                             .resize_image(image_file=file,
                                           resize_percent=THUMBNAIL_IMAGE_RESIZE_PERCENT),
                             compact=image_processing_service
                             .resize_image(image_file=file,
                                           resize_percent=COMPACT_IMAGE_RESIZE_PERCENT))
        else:
            instance = Image(name=name,
                             original=file,
                             thumbnail=file,
                             compact=image_processing_service
                             .resize_image(image_file=file,
                                           resize_percent=THUMBNAIL_IMAGE_RESIZE_PERCENT))

        return instance

    def bulk_create(self, files: list) -> list[Image]:
        images = [self.create(file=image, name=image.name) for image in files]
        return images

    def bulk_save(self, files: list) -> Iterator[Image]:
        new_images = self.bulk_create(files=files)
        return Image.objects.bulk_create(new_images)

    def delete(self, id: int):
        Image.objects.get(id=id).delete()


class DjangoImageReadRepository(ImageReadRepository):
    def all(self) -> Iterator[Image]:
        return Image.objects.all()

    def get(self, pk: int) -> Image:
        return Image.objects.get(pk=pk)

    def find(self, name: str) -> Iterator[Image]:
        return Image.objects.filter(name__icontains=name)



# Generated by Django 5.0.6 on 2024-06-05 13:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('simple_media_manager', '0002_image_compact_image_thumbnail_alter_image_image'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='image',
            options={'ordering': ('-created_at',)},
        ),
        migrations.RenameField(
            model_name='image',
            old_name='image',
            new_name='original',
        ),
    ]



from rest_framework import serializers

from simple_media_manager.infrastructure.common.serializers.custom_fields.fields import Base64ImageField
from simple_media_manager.infrastructure.models import Image


class ImageInputSerializer(serializers.Serializer):
    image = Base64ImageField(required=False, allow_null=False, allow_empty_file=False)
    name = serializers.CharField(required=False, allow_null=False)
    categories = serializers.ListField(required=True, allow_null=False, child=serializers.IntegerField())


class ImageMultipleInputSerializer(serializers.Serializer):
    images = serializers.ListField(child=serializers.ImageField(required=True, use_url=False))


class ImageOutputSerializer(serializers.ModelSerializer):
    class Meta:
        model = Image
        fields = "__all__"


class BriefImageOutputSerializer(serializers.ModelSerializer):
    class Meta:
        model = Image
        fields = ('id', 'thumbnail')


class SearchImageInputParameterSerializer(serializers.Serializer):
    name = serializers.CharField(required=False, allow_null=False)



from simple_media_manager.domain.repository.image import CompoundImageRepository
from simple_media_manager.infrastructure.adapters.image import DjangoImageReadRepository, DjangoImageWriteRepository

django_compound_image_repository = CompoundImageRepository(read_repository=DjangoImageReadRepository(),
                                                           write_repository=DjangoImageWriteRepository())



from io import BytesIO

from PIL import Image as PillowImage
from django.core.files import File as DjangoFile
from django.db import models, transaction

from ..models.file import File


class Image(File):
    original = models.ImageField(upload_to='media/images/originals', null=True, verbose_name='original size')
    thumbnail = models.ImageField(upload_to='media/images/thumbnails', null=True, verbose_name='thumbnail size')
    compact = models.ImageField(upload_to='media/images/compacts', null=True, verbose_name='compact size')

    def delete_files(self):
        self.original.delete()
        self.thumbnail.delete()
        self.compact.delete()

    def delete(self, using=None, keep_parents=False):
        with transaction.atomic():
            self.delete_files()
            super(Image, self).delete(using=using, keep_parents=keep_parents)

    def __str__(self):
        if not self.name:
            return self.original.name
        return self.name

    class Meta:
        ordering = ('-created_at',)
